<?php

declare(strict_types=1);

namespace Cin\AjaxSearch5;

use JTL\Plugin\Helper as PluginHelper;
use JTL\Plugin\PluginInterface;
use PDO;
use SimpleXMLElement;

final class Plugin
{
  private string $pluginid;

  private static PDO $pdo;

  private PluginInterface $oPlugin;

  public function __construct(?string $pluginid = null)
  {
    $infoxml = null === $pluginid
      ? (new SimpleXMLElement(file_get_contents(\dirname(__DIR__, 1).'/info.xml')))
      : (new SimpleXMLElement(file_get_contents(\dirname(__DIR__, 2)."/{$pluginid}/info.xml")));
    $this->pluginid = (string) $infoxml->PluginID[0];
  }

  /**
   * return boolean for checkboxes.
   */
  public function getConfig(string $key): mixed
  {
    $val = $this->getJTlPlugin()->getConfig()->getOption($key);
    if (null === $val) {
      return null;
    }

    if ('checkbox' === $val->inputType) {
      return 'on' === $val->value || '1' === $val->value;
    }

    return $val->value;
  }

  public function getPDO(): PDO
  {
    if (empty(self::$pdo)) {
      if ( ! \defined('DB_NAME')) {
        require \dirname(__DIR__, 3).'/includes/config.JTL-Shop.ini.php';
      }
      self::$pdo = new PDO(
        'mysql:dbname='.DB_NAME.';host='.DB_HOST.';charset=utf8mb4',
        DB_USER,
        DB_PASS,
        [
          PDO::ATTR_DEFAULT_FETCH_MODE => PDO::FETCH_OBJ,
          PDO::ATTR_ERRMODE            => PDO::ERRMODE_EXCEPTION,
          PDO::ATTR_EMULATE_PREPARES   => false,
          PDO::ATTR_STRINGIFY_FETCHES  => false
        ]
      );
    }

    return self::$pdo;
  }

  public function getJTlPlugin(): PluginInterface
  {
    if ( ! class_exists('JTL\Plugin\Helper')) {
      require_once \dirname(__DIR__, 3).'/includes/globalinclude.php';
    }
    if (empty($this->oPlugin)) {
      $id            = PluginHelper::getIDByPluginID($this->pluginid);
      $this->oPlugin = PluginHelper::getLoaderByPluginID($id)->init($id);
    }

    return $this->oPlugin;
  }

  public function getPluginID(): string
  {
    return $this->pluginid;
  }

  public function lang(): array
  {
    return $this->getJTlPlugin()->getLocalization()->getTranslations();
  }

  public function cfg(): array
  {
    $configPath        = $this->getJTlPlugin()->getPaths()->getBasePath().'/common/settings/';
    $defaultConfigFile = $configPath.'default.php';
    $customConfigFile  = $configPath.'custom.php';

    $cfg = [];

    if (file_exists($defaultConfigFile)) {
      require $defaultConfigFile;
      foreach ($defaultConfig as $name => $value) {
        $cfg[$name] = $value;
      }
    }

    foreach ($this->getJTlPlugin()->getConfig()->getAssoc() as $name => $option) {
      $cfg[$name] = $option->value;
    }

    if (isset($defaultConfig) && file_exists($customConfigFile)) {
      require $customConfigFile;
      foreach ($customConfig as $name => $value) {
        if (\in_array($name, array_keys($cfg), true)) {
          $cfg[$name] = $value;
        }
      }
    }

    return $cfg;
  }

  public function thirdPartyLibraries(): array
  {
    $partys = [];

    $composerFilename   = \dirname(__DIR__, 1).'/composer.lock';
    $partys['composer'] = file_exists($composerFilename) ? file_get_contents($composerFilename) : '{}';

    $thirdPartyFilename   = \dirname(__DIR__, 1).'/thirdparty.json';
    $partys['thirdparty'] = file_exists($thirdPartyFilename) ? file_get_contents($thirdPartyFilename) : '{}';

    return array_merge(json_decode($partys['composer'])->packages ?? [], json_decode($partys['thirdparty'])->packages ?? []);
  }

  public function isForeignPluginInstalled(string $pluginid): bool
  {
    $stmt = $this->getPDO()->prepare('SELECT COUNT(*)
      FROM `tplugin`
      WHERE cPluginID = :pluginid AND nStatus = 2
    ');
    $stmt->execute([
      ':pluginid' => $pluginid
    ]);

    return 1 === $stmt->fetchColumn();
  }
}
